@@ -1,15 +1,20 @@
+"""
+Scrapping module for fetching data from B3's APIs.
+"""
+
 import base64
 from concurrent.futures import ThreadPoolExecutor
-from datetime import date
+from datetime import date, datetime
 import logging
-from io import StringIO
+from io import BytesIO, StringIO
 from typing import Iterable, Union, Any
 
 import requests
 import pandas as pd
+import numpy as np
 
-from b3od.exceptions import RequestException
-from b3od.utils import parse_date
+from b3od.utils import parse_date, is_sequence
+from b3od.meta import SERVICES_DTYPES, SERVICES_DATE_COLUMNS
 
 logger = logging.getLogger(__name__)
 
@@ -33,7 +38,7 @@ class B3Scrapper:
 
     web_consolidated_base_url = "https://arquivos.b3.com.br/api"
 
-    bmfbovespa_base_url = ""
+    bmfbovespa_base_url = "https://bvmf.bmfbovespa.com.br"
 
     def __init__(self, session: requests.Session = None):
         self.session = session
@@ -59,44 +64,57 @@ class B3Scrapper:
         """List all available services"""
         return list(self.SERVICE_DICT.keys())
 
-    def _basic_request(
-        self, method: str, url: str, data: dict = None
-    ) -> requests.Response:
+    def __basic_request(self, method: str, url: str, **kwargs) -> requests.Response:
         if self.session is None:
             self.start()
 
-        if data is None:
-            data = {}
-
         logger.debug("Sending '%s' request to '%s'", method, url)
-        response = self.session.request(method, url, data=data)
+        response = self.session.request(method, url, **kwargs)
 
         logger.debug("Response status code: %s", response.status_code)
         if response.status_code != 200:
-            raise RequestException(response)
+            raise response.raise_for_status()
 
         logger.debug("Response content: %s", response.text[:100] + "...")
         return response
 
-    def __download_web_consolidated(self, table: str, dt: Any) -> pd.DataFrame:
+    def __download_web_consolidated(self, table: str, query_date: Any) -> pd.DataFrame:
         """Downloads data from a table in B3 website for a given date"""
 
-        dt = parse_date(dt)
+        query_date = parse_date(query_date)
 
-        url = f"{self.web_consolidated_base_url}/download/requestname?fileName={table}&date={dt:%Y-%m-%d}&recaptchaToken="
-        response = self._basic_request("get", url)
+        url = f"{self.web_consolidated_base_url}/download/requestname"
+
+        params = {
+            "fileName": table,
+            "date": query_date.strftime("%Y-%m-%d"),
+            "recaptchaToken": "",
+        }
+
+        response = self.__basic_request("get", url, params=params)
 
         redirect_url = response.json()["redirectUrl"].replace(
             "~", self.web_consolidated_base_url
         )
-        response = self._basic_request("get", redirect_url)
+        response = self.__basic_request("get", redirect_url)
 
         if response.text == "":
-            logger.error("No data found for %s", dt)
+            logger.error("No data found for %s", query_date)
             return pd.DataFrame()
 
-        df = pd.read_csv(StringIO(response.text), sep=";", decimal=",")
-        return df
+        queried_data = pd.read_csv(
+            StringIO(response.text),
+            sep=";",
+            decimal=",",
+            dtype=SERVICES_DTYPES[table],
+            parse_dates=SERVICES_DATE_COLUMNS[table],
+            date_parser=(
+                lambda x: datetime.strptime(x, "%Y-%m-%d").date()
+                if not pd.isna(x)
+                else np.nan
+            ),
+        )
+        return queried_data
 
     def __download_web_consolidated_many(
         self, table: str, dts: Iterable[Any]
@@ -110,16 +128,28 @@ class B3Scrapper:
             ]
             return pd.concat([f.result() for f in futures], ignore_index=True)
 
-    def __download_position_limits(self, dt: date) -> pd.DataFrame:
+    def __download_position_limits(self, query_date: date) -> pd.DataFrame:
+        table = "PositionLimits"
+        query_date = parse_date(query_date)
+        url = f"{self.bmfbovespa_base_url}/LimitesPosicoes/Posicoes/DownloadArquivoDiretorio"
 
-        dt = parse_date(dt)
-        url = f"https://bvmf.bmfbovespa.com.br/LimitesPosicoes/Posicoes/DownloadArquivoDiretorio?data={dt:%Y%m%d}"
-        response = self._basic_request("get", url)
+        params = {"data": query_date.strftime("%Y-%m-%d")}
 
-        df = pd.read_csv(
-            StringIO(base64.b64decode(response.text)), sep=";", decimal=","
+        response = self.__basic_request("get", url, params=params)
+
+        queried_data = pd.read_csv(
+            BytesIO(base64.b64decode(response.text)),
+            sep=";",
+            decimal=",",
+            dtype=SERVICES_DTYPES[table],
+            parse_dates=SERVICES_DATE_COLUMNS[table],
+            date_parser=(
+                lambda x: datetime.strptime(x, "%Y-%m-%d").date()
+                if not pd.isna(x)
+                else np.nan
+            ),
         )
-        return df
+        return queried_data
 
     def __download_position_limits_many(
         self, table: str, dts: Iterable[Any]
@@ -133,18 +163,22 @@ class B3Scrapper:
             ]
             return pd.concat([f.result() for f in futures], ignore_index=True)
 
-    def download(self, table: str, dt: Union[Any, Iterable[Any]]) -> pd.DataFrame:
+    def download(
+        self, table: str, query_date: Union[Any, Iterable[Any]]
+    ) -> pd.DataFrame:
         """Downloads data from a table in B3 website for a given date (or set of dates)"""
         if not table in self.SERVICE_DICT:
             raise ValueError(f"Table must be one of {self.SERVICE_DICT.keys()}")
         service = self.SERVICE_DICT[table]
 
         if service == "WebConsolidated":
-            if isinstance(dt, Iterable):
-                return self.__download_web_consolidated_many(table, dt)
-            return self.__download_web_consolidated(table, dt)
+            if is_sequence(query_date):
+                print("Downloading many")
+                return self.__download_web_consolidated_many(table, query_date)
+            print("Downloading one")
+            return self.__download_web_consolidated(table, query_date)
         if service == "PositionLimits":
-            if isinstance(dt, Iterable):
-                return self.__download_position_limits_many(table, dt)
-            return self.__download_position_limits(dt)
+            if is_sequence(query_date):
+                return self.__download_position_limits_many(table, query_date)
+            return self.__download_position_limits(query_date)
         raise ValueError(f"Table must be one of {set(self.SERVICE_DICT.values())}")
